/*
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef AVCODEC_LCEVC_PARSE_H
#define AVCODEC_LCEVC_PARSE_H

#include <stdint.h>

#include "get_bits.h"

static inline uint64_t get_mb(GetBitContext *s) {
    int more, i = 0;
    uint64_t mb = 0;

    do {
        int byte = get_bits(s, 8);
        unsigned bits = byte & 0x7f;
        more = byte & 0x80;
        mb = (mb << 7) | bits;
        if (++i == 10)
            break;
    } while (more);

    return mb;
}

#endif /* AVCODEC_LCEVC_PARSE_H */
